(*
  LibOVR wrapper for Delphi

  Copyright (c) 2014 Lyna

  This software is provided 'as-is', without any express or implied
  warranty. In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software
    in a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

    3. This notice may not be removed or altered from any source
    distribution.
*)
unit LibOVR.GL;

interface

uses
  Winapi.Windows, Winapi.OpenGL, System.SysUtils, LibOVR;

type
  ovrGLConfigData = record
    /// General device settings.
    Header: ovrRenderAPIConfigHeader;

    /// The optional window handle. If unset, rendering will use the current window.
    Window: HWND;
    /// The optional device context. If unset, rendering will use a new context.
    DC: HDC;
  end;

  /// Contains OpenGL-specific rendering information.
  ovrGLConfig = record
    case Integer of
      /// General device settings.
      0: (Config: ovrRenderAPIConfig);
      /// OpenGL-specific settings.
      1: (OGL: ovrGLConfigData);
  end;

  /// Used to pass GL eye texture data to ovrHmd_EndFrame.
  ovrGLTextureData = record
    /// General device settings.
    Header: ovrTextureHeader;
    /// The OpenGL name for this texture.
    TexId: GLuint;
  end;

  /// Contains OpenGL-specific texture information.
  ovrGLTexture = record
    case Integer of
      /// General device settings.
      0: (Texture: ovrTexture);
      /// OpenGL-specific settings.
      1: (OGL: ovrGLTextureData);
  end;

  { For wrapper class }

  TeyeRenderViewport = array[ovrEye_Left..ovrEye_Right] of ovrRecti;

  THmdCap = (
    // Read-only
    hcPresent, hcAvailable, hcCaptured, hcExtendDesktop,
    // Modifiable
    hcNoMirrorToWindow, hcDisplayOff, hcLowPersistence, hcDynamicPrediction, hcNoVSync);
  THmdCaps = set of THmdCap;

  TDistortionCap = (dcChromatic, dcTimeWarp, dcVignette, dcNoRestore,
    dcFlipInput, dcSRGB, dcOverdrive, dcHqDistortion, dcProfileNoTimewarpSpinWaits);
  TDistortionCaps = set of TDistortionCap;

  TTrackingCap = (tcOrientation, tcMagYawCorrection, tcPosition, tcIdle);
  TTrackingCaps = set of TTrackingCap;

  TRenderFrameProc = reference to procedure(const APose: ovrPosef;
    const AFov: ovrFovPort; const AViewAdjust: ovrVector3f; const AViewport: ovrRecti);

  TOculusRift = class
  private
    FHandle: ovrHmd;
    FWindowHandle: THandle;
    FEyeRenderDesc: TeyeRenderDescOut;
    FEyeRenderViewport: TeyeRenderViewport;
    FEyeTexture: TeyeTexture;
    FEyeRenderPose: TrenderPose;
    class function GetAvailable: Boolean; static;
    class constructor Create;
    class destructor Destroy;
  public
    constructor Create(AWindowHandle: THandle);
    constructor CreateDebug(AHmdType: ovrHmdType; AWindowHandle: THandle);
    destructor Destroy; override;
    procedure InitializeOpenGL(ATexture: Integer);
    procedure Resize(const AWidth, AHeight: Integer);

    function GetFovTextureSize(AEye: ovrEyeType): ovrSizei;
    function GetRenderTargetSize: ovrSizei;
    procedure SetEnabledCaps(const AHmdCaps: THmdCaps);
    function ConfigureRendering(const AConfig: ovrRenderAPIConfig;
      const ADistortionCaps: TDistortionCaps): Boolean;
    function ConfigureTracking(const ATrackingCap: TTrackingCaps): Boolean;
    procedure BeginFrame;
    procedure EndFrame;
    procedure RenderFrame(const AProc: TRenderFrameProc);
    procedure DismissHSWDisplay;
    procedure RecenterPose;

    class property Available: Boolean read GetAvailable;
    property Handle: ovrHmd read FHandle;
    property EyeRenderViewport: TeyeRenderViewport read FEyeRenderViewport;
    property EyeRenderPose: TrenderPose read FEyeRenderPose;
  end;

implementation

uses
  System.Math;

{ TOculusRift }

class function TOculusRift.GetAvailable: Boolean;
begin
  Result := ovrHmd_Detect > 0;
end;

class constructor TOculusRift.Create;
begin
  ovr_Initialize;
end;

class destructor TOculusRift.Destroy;
begin
  ovr_Shutdown;
end;

constructor TOculusRift.Create(AWindowHandle: THandle);
begin
  inherited Create;
  FHandle := ovrHmd_Create(0);
  if FHandle = nil then
    raise Exception.Create('Failed to create handle.');

  FWindowHandle := AWindowHandle;
  ovrHmd_AttachToWindow(FHandle, FWindowHandle);
end;

constructor TOculusRift.CreateDebug(AHmdType: ovrHmdType; AWindowHandle: THandle);
begin
  inherited Create;
  FHandle := ovrHmd_CreateDebug(ovrHmd_DK2);
  if FHandle = nil then
    raise Exception.Create('Failed to create handle.');

  FWindowHandle := AWindowHandle;
  ovrHmd_AttachToWindow(FHandle, FWindowHandle);
end;

destructor TOculusRift.Destroy;
begin
  if FHandle <> nil then
    ovrHmd_Destroy(FHandle);
  inherited;
end;

procedure TOculusRift.InitializeOpenGL(ATexture: Integer);
var
  renderTargetSize: ovrSizei;
begin
  renderTargetSize := GetRenderTargetSize;

  FEyeRenderViewport[ovrEye_Left].Pos.x := 0;
  FEyeRenderViewport[ovrEye_Left].Pos.y := 0;
  FEyeRenderViewport[ovrEye_Left].Size.w := renderTargetSize.w div 2;
  FEyeRenderViewport[ovrEye_Left].Size.h := renderTargetSize.h;
  FEyeRenderViewport[ovrEye_Right].Pos.x := (renderTargetSize.w + 1) div 2;
  FEyeRenderViewport[ovrEye_Right].Pos.y := 0;
  FEyeRenderViewport[ovrEye_Right].Size := FEyeRenderViewport[ovrEye_Left].Size;

  FEyeTexture[0].Header.API := ovrRenderAPI_OpenGL;
  FEyeTexture[0].Header.TextureSize := renderTargetSize;
  FEyeTexture[0].Header.RenderViewport := FEyeRenderViewport[ovrEye_Left];
  FEyeTexture[0].PlatformData[0] := ATexture;
  FEyeTexture[1] := FEyeTexture[0];
  FEyeTexture[1].Header.RenderViewport := FEyeRenderViewport[ovrEye_Right];

  SetEnabledCaps([hcLowPersistence, hcDynamicPrediction]);
  if not ConfigureTracking([tcOrientation, tcMagYawCorrection, tcPosition]) then
    raise Exception.Create('Failed to configure tracking parameters.');
end;

procedure TOculusRift.Resize(const AWidth, AHeight: Integer);
var
  config: ovrGLConfig;
begin
  config.OGL.Header.API := ovrRenderAPI_OpenGL;
  config.OGL.Header.RTSize.w := AWidth;
  config.OGL.Header.RTSize.h := AHeight;
  config.OGL.Header.Multisample := 1;
  config.OGL.Window := FWindowHandle;
  config.OGL.DC := 0;
  if not ConfigureRendering(config.Config, [dcChromatic, dcVignette, dcTimeWarp, dcOverdrive, dcHqDistortion]) then
    raise Exception.Create('Failed to configure rendering parameters.');
end;

function TOculusRift.GetFovTextureSize(AEye: ovrEyeType): ovrSizei;
begin
  Result := ovrHmd_GetFovTextureSize(FHandle, AEye, FHandle^.DefaultEyeFov[Ord(AEye)], 1.0);
end;

function TOculusRift.GetRenderTargetSize: ovrSizei;
var
  recommenedTex0Size, recommenedTex1Size: ovrSizei;
begin
  recommenedTex0Size := GetFovTextureSize(ovrEye_Left);
  recommenedTex1Size := GetFovTextureSize(ovrEye_Right);
  Result.w := recommenedTex0Size.w + recommenedTex1Size.w;
  Result.h := Max(recommenedTex0Size.h, recommenedTex1Size.h);
end;

procedure TOculusRift.SetEnabledCaps(const AHmdCaps: THmdCaps);
const
  HmdCapValues: array[THmdCap] of Longword = (
    ovrHmdCap_Present,
    ovrHmdCap_Available,
    ovrHmdCap_Captured,
    ovrHmdCap_ExtendDesktop,
    ovrHmdCap_NoMirrorToWindow,
    ovrHmdCap_DisplayOff,
    ovrHmdCap_LowPersistence,
    ovrHmdCap_DynamicPrediction,
    ovrHmdCap_NoVSync
  );
var
  caps: ovrHmdCaps;
  cap: THmdCap;
begin
  caps := 0;
  for cap := Low(THmdCap) to High(THmdCap) do
    if cap in AHmdCaps then
      caps := caps or HmdCapValues[cap];

  ovrHmd_SetEnabledCaps(FHandle, caps);
end;

function TOculusRift.ConfigureRendering(const AConfig: ovrRenderAPIConfig;
  const ADistortionCaps: TDistortionCaps): Boolean;
const
  DistortionCapValues: array[TDistortionCap] of Longword = (
    ovrDistortionCap_Chromatic,
    ovrDistortionCap_TimeWarp,
    ovrDistortionCap_Vignette,
    ovrDistortionCap_NoRestore,
    ovrDistortionCap_FlipInput,
    ovrDistortionCap_SRGB,
    ovrDistortionCap_Overdrive,
    ovrDistortionCap_HqDistortion,
    ovrDistortionCap_ProfileNoTimewarpSpinWaits
  );
var
  caps: ovrDistortionCaps;
  cap: TDistortionCap;
  eyeFovIn: TeyeFovIn;
begin
  caps := 0;
  for cap := Low(TDistortionCap) to High(TDistortionCap) do
    if cap in ADistortionCaps then
      caps := caps or DistortionCapValues[cap];

  eyeFovIn[0] := FHandle^.DefaultEyeFov[0];
  eyeFovIn[1] := FHandle^.DefaultEyeFov[1];

  Result := ovrHmd_ConfigureRendering(FHandle, AConfig, caps, eyeFovIn, FEyeRenderDesc);
end;

function TOculusRift.ConfigureTracking(const ATrackingCap: TTrackingCaps): Boolean;
const
  TrackingCapValues: array[TTrackingCap] of Longword = (
    ovrTrackingCap_Orientation,
    ovrTrackingCap_MagYawCorrection,
    ovrTrackingCap_Position,
    ovrTrackingCap_Idle
  );
var
  caps: ovrTrackingCaps;
  cap: TTrackingCap;
begin
  caps := 0;
  for cap := Low(TTrackingCap) to High(TTrackingCap) do
    if cap in ATrackingCap then
      caps := caps or TrackingCapValues[cap];

  Result := ovrHmd_ConfigureTracking(FHandle, caps, 0);
end;

procedure TOculusRift.BeginFrame;
begin
  ovrHmd_BeginFrame(FHandle, 0);
end;

procedure TOculusRift.EndFrame;
begin
  ovrHmd_EndFrame(FHandle, FEyeRenderPose, FEyeTexture);
end;

procedure TOculusRift.RenderFrame(const AProc: TRenderFrameProc);
var
  eyeIndex: Integer;
  eye: ovrEyeType;
begin
  for eyeIndex := 0 to Ord(ovrEye_Count)-1 do
  begin
    eye := FHandle^.EyeRenderOrder[eyeIndex];
    FEyeRenderPose[eye] := ovrHmd_GetEyePose(FHandle, eye);
    AProc(FEyeRenderPose[eye], FEyeRenderDesc[eyeIndex].Fov,
      FEyeRenderDesc[eyeIndex].ViewAdjust, FEyeRenderViewport[eye]);
  end;
end;

procedure TOculusRift.DismissHSWDisplay;
begin
  ovrHmd_DismissHSWDisplay(FHandle);
end;

procedure TOculusRift.RecenterPose;
begin
  ovrHmd_RecenterPose(FHandle);
end;

end.
